"""
Setup of the slack notification.  The default slack notifier that is provided by Prefect had two issues that
needed fixing:

    1) The URL of the link was wrong
    2) No rate limiting

"""
# pylint: disable=broad-except
# pylint: disable=no-name-in-module
# pylint: disable=import-error
# pylint: disable=no-member
# pylint: disable=too-many-branches

# Import standard libraries
import datetime
import json
import socket
from io import BytesIO
from typing import cast

# Related 3rd party imports
import azure.core.exceptions
import prefect
import requests
from prefect.engine.state import Failed, TimedOut, TriggerFailed
from prefect.utilities.graphql import with_args

from .storage import AzureStorage

DEBUG = False  # if set to True, notifications will also be sent when running locally

from .logger import logger

NOTIFICATION_FILEBUCKET = "runinfo"
NOTIFICATION_FILEPATH = "prefect/state/notification_state.json"
NOTIFICATION_RATE_LIMIT = (
    7200  # only allow 1 notification for every 7200 seconds (per flow)
)


def customized_slack_notifier(config, flow, new_state, _reference_task_states):
    """
    Customized version of the slack notifier.  Only send out notifications for a specific flow at most
    once every 2 hours.
    """
    webhook_url = config.get("SLACK_WEBHOOK_URL")
    if webhook_url is None:
        if DEBUG:
            # fallback to posting to '#pedgar-test'
            webhook_url = "https://hooks.slack.com/services/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"
        else:
            logger.warning(
                "No slack webhook found.  Not sending any failure notification."
            )
            return new_state
    send_notification = False
    if isinstance(new_state, Failed):
        flow_name = prefect.context.get("flow_name")
        notifications = get_notification_state(config)

        if flow_name in notifications:
            last_notified = datetime.datetime.fromisoformat(notifications[flow_name])
            if last_notified < datetime.datetime.now() - datetime.timedelta(
                seconds=NOTIFICATION_RATE_LIMIT
            ):
                send_notification = True
            else:
                logger.info(
                    "Last notified about '%s' at %s.  Not sending new notification for at least %s seconds",
                    flow_name,
                    last_notified,
                    (
                        last_notified
                        + datetime.timedelta(seconds=NOTIFICATION_RATE_LIMIT)
                        - datetime.datetime.now()
                    ).total_seconds(),
                )
        else:
            send_notification = True
            logger.info("Don't know the last time we notified about '%s'", flow_name)

        if send_notification:
            # Don't send notifications when testing locally
            if DEBUG or not is_running_locally():
                notifications[flow_name] = datetime.datetime.now().isoformat()
                set_notification_state(config, notifications)

                form_data = get_task_runs_for_flow_local(flow, new_state)
                if form_data is None:
                    form_data = get_task_runs_for_flow_cloud(flow, new_state)

                resp = requests.post(webhook_url, json=form_data)
                if not resp.ok:
                    raise ValueError("Slack notification for {} failed".format(flow))
            else:
                logger.info(
                    "Not sending slack notification because we're running locally and DEBUG isn't enabled"
                )
    return new_state


def get_task_runs_for_flow_cloud(flow, _new_state):
    """
    Get the list of task runs when the Prefect flow is run on the Prefect server (not locally)

    Uses the GraphQL API to find out the information because of the following bug:
        https://github.com/PrefectHQ/prefect/issues/4570
    """
    flow_run_id = prefect.context.get("flow_run_id")
    query_flows = {
        "query": {
            with_args(
                "task_run",
                {
                    "where": {
                        "flow_run_id": {"_eq": flow_run_id},
                    },
                },
            ): {
                "id": True,
                "name": True,
                "task": {
                    "name": True,
                },
                "state": True,
                "state_message": True,
            }
        }
    }

    prefect_client = prefect.client.Client()
    task_runs = prefect_client.graphql(query_flows).data.task_run
    failed_task = None
    failed_state = None
    failed_state_message = None
    for task_info in task_runs:
        if task_info["state"] == "Failed":
            failed_task = task_info["task"]["name"]
            failed_state = task_info["state"]
            failed_state_message = task_info["state_message"]
            break
        if task_info["state"] in ["TriggerFailed", "TimedOut"] and failed_task is None:
            failed_task = task_info["task"]["name"]
            failed_state = task_info["state"]
            failed_state_message = task_info["state_message"]

    if failed_task is not None:
        return slack_message_formatter(
            flow.name, failed_task, failed_state, failed_state_message
        )

    logger.warning(
        "Failed to find failed task info for flow %s. 'task_runs' = %s",
        flow.name,
        task_runs,
    )
    return None


def get_task_runs_for_flow_local(flow, new_state):
    """
    Get the list of task runs when the Prefect flow is run locally (not on the Prefect server)

    Uses a different method to find out the information because of the following bug:
        https://github.com/PrefectHQ/prefect/issues/4570
    """
    failed_task = None
    failed_state = None

    for task, task_state in new_state.result.items():
        if isinstance(task_state, Failed):
            if not isinstance(task_state, TriggerFailed) and not isinstance(
                task_state, TimedOut
            ):
                failed_task = task
                failed_state = task_state
                break
            if failed_task is None:
                failed_task = task
                failed_state = task_state

    if failed_task is not None:
        return slack_message_formatter_from_obj(flow, failed_task, failed_state)
    return None


def get_notification_state(config):
    """
    Get the list of flows and the last time a notification was sent for them
    """
    notifications = {}
    try:
        data_text = AzureStorage(
            config.get("AZURE_STORAGE_ACCOUNT_NAME"),
            config.get("AZURE_STORAGE_ACCOUNT_KEY"),
        ).read(
            NOTIFICATION_FILEBUCKET,
            NOTIFICATION_FILEPATH,
        )
        notifications = json.loads(data_text)
    except azure.core.exceptions.ResourceNotFoundError:
        pass
    except Exception:
        logger.exception("got an unexpected exception")
    return notifications


def set_notification_state(config, notifications):
    """
    Store the list of flows and the last time a notification was sent for them
    """
    try:
        AzureStorage(
            config.get("AZURE_STORAGE_ACCOUNT_NAME"),
            config.get("AZURE_STORAGE_ACCOUNT_KEY"),
        ).store(
            NOTIFICATION_FILEBUCKET,
            NOTIFICATION_FILEPATH,
            BytesIO(str.encode(json.dumps(notifications))),
            overwrite=True,
        )
    except Exception:
        logger.exception("Got an unexpected exception while storing notification state")


def is_running_locally():
    """
    Return False if running remotely on the Prefect server, True otherwise.

    So far, the only known way to determine this is by looking at the local hostname.  If running on the prefect
    server, the local hostname will look like:
        prefect-job-b5837576-2zlhs
        prefect-job-4e5eca28-zp4t9
    """
    return "prefect" not in socket.getfqdn()


def slack_message_formatter_from_obj(
    flow,
    task,
    state: "prefect.engine.state.State",
) -> dict:
    """Return a Slack message block"""
    state_name = type(state).__name__
    # flow_name = prefect.context.get("flow_name")
    if isinstance(state.result, Exception):
        state_message = repr(state.result)
    else:
        state_message = cast(str, state.message)
    return slack_message_formatter(flow.name, task.name, state_name, state_message)


def slack_message_formatter(
    flow_name: str, task_name: str, state_name: str, state_message: str
) -> dict:
    """Return a Slack message block"""
    fields = []
    value = "```{}```".format(state_message)
    fields.append({"title": "Message", "value": value, "short": False})

    url = None
    title = state_name
    if prefect.context.get("flow_run_id") and not is_running_locally():
        url = f"http://prefect.apps.optimeering.hawkai.site/default/flow-run/{prefect.context['flow_run_id']}"
        title += " (click for logs)"
        # notification_payload.update(title_link=url)

    notification_payload = {
        "fallback": f"{flow_name} failed",
        "color": "#eb0000",
        "author_name": "Prefect",
        "author_link": "https://www.prefect.io/",
        "author_icon": "https://emoji.slack-edge.com/TAN3D79AL/prefect/2497370f58500a5a.png",
        "title": title,
        "fields": fields,
        "title_link": url,
        "text": f"_{flow_name}_: '{task_name}' is now in a {state_name} state",
        "footer": "Prefect notification",
    }

    data = {"attachments": [notification_payload]}
    return data



import os
import shlex
import subprocess
from numbers import Number
from typing import Dict

import git
import prefect
from prefect.executors import DaskExecutor
from prefect.run_configs import KubernetesRun
from prefect.storage import Docker
from prefect.utilities.graphql import with_args

from .logger import logger


class DockerBaseImage:
    """Helper class for building and pushing base docker image for use with prefect"""

    def __init__(
        self,
        az_registry_url,
        project_base_path,
        base_docker_image,
        skip_login=False,
    ):
        self.az_registry_url = az_registry_url
        self.project_base_path = project_base_path
        self.base_docker_image = base_docker_image
        self.skip_login = skip_login

    def build_and_push(self):
        if not self.skip_login:
            self.login()
        self.pull()
        self.build()
        self.push()

    def login(self):
        az_docker_registry_name = self.az_registry_url.split(".")[0]
        logger.info(
            f"Logging into Azure (running `az acr login --name {az_docker_registry_name}`)"
        )
        subprocess.run(
            f"az acr login --name {az_docker_registry_name}".split(), check=True
        )

    def pull(self):
        logger.info("Pulling existing base image")
        try:
            subprocess.run(
                [
                    "docker",
                    "pull",
                    self.base_docker_image,
                ],
                check=True,
            )
        except subprocess.CalledProcessError:
            # We're pulling it only because if there is a recent build, the old layers
            # can be re-used. So, it is not critical if this failed.
            logger.warn("Failed to pull existing base image. Continuing anyway...")

    def build(self):
        logger.info("Building base image")

        run_cmd = [
            "docker",
            "buildx",
            "build",
            "--platform",
            "linux/amd64",
            "--pull",
            "-t",
            self.base_docker_image,
            "-f",
            os.path.join(self.project_base_path, "Dockerfile"),
            self.project_base_path,
        ]

        run_cmd = " ".join([shlex.quote(run_arg) for run_arg in run_cmd])

        subprocess.check_output(
            run_cmd,
            shell=True,
        )

    def push(self):
        logger.info("Uploading base image")
        subprocess.run(
            [
                "docker",
                "push",
                self.base_docker_image,
            ],
            check=True,
        )


def create_docker_storage(
    project_base_path: str,
    az_docker_registry_url: str,
    az_docker_image_base_name: str,
    env_vars: Dict,
    skip_login: bool = False,
) -> prefect.storage.Docker:
    """
    Build docker image to be used as a base for deploying prefect flow. Only supports
    Azure Container Service.

    Args:
        project_base_path (str):
            Docker build direction; the Dockerfile is expected in this path.
        az_docker_registry_url (str):
            Azure registry url to be used for the docker image.
        az_docker_image_base_name (str):
            Image name to be used as the base image.
        env_vars (typing.Dict):
            Dictionary of environment variables to be embedded in the docker image.
        skip_login (bool):
            Skip logging in to docker (assumes that it is already logged in by other
            means).

    Returns:
        prefect.storage.Docker:
            prefect Docker storage object to be used with prefect flow.
    """

    base_docker_image = f"{az_docker_registry_url}/{az_docker_image_base_name}"
    DockerBaseImage(
        az_docker_registry_url,
        project_base_path,
        base_docker_image,
        skip_login,
    ).build_and_push()

    git_info = _get_git_info()
    env_vars = {**env_vars, **git_info}

    return Docker(
        registry_url=az_docker_registry_url,
        env_vars=env_vars,
        base_image=base_docker_image,
    )


def deploy_flow(
    flow: prefect.Flow,
    schedule: prefect.schedules.Schedule,
    docker_storage: prefect.storage.Docker,
    prefect_project_name: str,
    mem_limit_gb: Number = None,
) -> None:
    """Deploy the Prefect flow

    Build and deploy a given prefect flow.

    Args:
        flow (prefect.Flow):
            prefect flow to deploy
        schedule (prefect.schedules.Schedule):
            schedule to run the deployed flow
        docker_storage (prefect.storage.Docker):
            docker storage to store the flow to
        prefect_project_name (typing.Dict):
            prefect project name to deploy the flow to
        mem_limit_gb (Number, optional):
            override the default memory limit for docker image in the dask cluster
    """

    flow.storage = docker_storage

    if schedule:
        flow.schedule = schedule

    flow.run_config = KubernetesRun()

    if mem_limit_gb:
        flow.executor = DaskExecutor(
            cluster_kwargs=dict(memory_limit="%sGi" % mem_limit_gb), debug=True
        )
    else:
        flow.executor = DaskExecutor(debug=True)

    create_project_if_not_exist(prefect_project_name)

    flow.register(project_name=prefect_project_name)


def create_project_if_not_exist(project_name, prefect_client=None):
    """Create prefect project if it doesn't already exist

    project_name is mandatory in prefect version 0.13.0 and higher when registering a
    flow. So, it needs to be created.
    """

    if prefect_client is None:
        prefect_client = prefect.client.Client()

    query_project = {
        "query": {
            with_args("project", {"where": {"name": {"_eq": project_name}}}): {
                "id": True
            }
        }
    }

    project = prefect_client.graphql(query_project).data.project
    if not project:
        prefect_client.create_project(project_name)


def _get_git_info():
    """Return the Git information"""
    repo = git.Repo(search_parent_directories=True)

    try:
        branch = str(repo.active_branch)
    except TypeError:
        branch = ""

    tags = [
        str(tag)
        for tag in repo.tags
        if tag is not None and tag.commit == repo.head.commit
    ]

    return {
        "GIT_COMMIT_ID": repo.head.commit.hexsha,
        "GIT_IS_DIRTY": str(repo.is_dirty()),
        "GIT_BRANCH": branch,
        "GIT_TAGS": ", ".join(tags),
    }



from functools import partial
from io import BytesIO

from azure.storage.blob import BlobClient

from .logger import logger


class AzureStorage:
    def __init__(self, azure_storage_account_name, azure_storage_account_key):
        self.blob_client = partial(
            BlobClient.from_connection_string,
            conn_str=self._get_azure_connection_string(
                azure_storage_account_name, azure_storage_account_key
            ),
            timeout=120,
        )

    def read(self, container_name, blob_name):
        download_stream = self.blob_client(
            container_name=container_name, blob_name=blob_name
        ).download_blob()
        data_text = download_stream.readall()
        return data_text

    def store(
        self,
        container_name: str,
        blob_name: str,
        contents: BytesIO,
        overwrite: bool = False,
    ):
        logger.info(
            f"Writing blob '{blob_name}' to container '{container_name}' ({contents.getbuffer().nbytes} bytes)"
        )

        file_info = self.blob_client(
            container_name=container_name, blob_name=blob_name
        ).upload_blob(
            data=contents,
            overwrite=overwrite,
        )

        etag = file_info["etag"]
        logger.info(f"Stored with etag {etag}")

    @staticmethod
    def _get_azure_connection_string(
        azure_storage_account_name, azure_storage_account_key
    ):
        """
        Return the connection string used to connect to Azure
        """

        return ";".join(
            [
                "DefaultEndpointsProtocol=https",
                f"AccountName={azure_storage_account_name}",
                f"AccountKey={azure_storage_account_key}",
                "EndpointSuffix=core.windows.net",
            ]
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['archimedes_flow_utils']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.27,<4.0.0',
 'azure-core>=1.23.1,<2.0.0',
 'azure-storage-blob>=12.11.0,<13.0.0',
 'prefect>=1.2.0,<2.0.0',
 'python-dotenv>=0.20.0,<0.21.0',
 'requests>=2.27.1,<3.0.0']

setup_kwargs = {
    'name': 'archimedes-flow-utils',
    'version': '1.1.6',
    'description': 'Common code for all our prefect flows',
    'long_description': 'None',
    'author': 'Optimeering AS',
    'author_email': 'dev@optimeering.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.12',
}


setup(**setup_kwargs)
